///File : main.cpp
//------------------------------------------------------------------------------
/*
    This file is part of Debtest.

    Debtest is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Debtest is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Debtest.  If not, see <http://www.gnu.org/licenses/>.
*/
//------------------------------------------------------------------------------
#include <iostream>
#include <vector>
#include <fstream>

#include <stdio.h>
#include <stdlib.h>

#include <pthread.h>
//------------------------------------------------------------------------------
#include "common.hpp"
#include "main.hpp"
#include "config.hpp"
//------------------------------------------------------------------------------
using namespace std;
//------------------------------------------------------------------------------
void checkFilesExistence(vector<errListStr> &errList, const vector<string> &pkgList, progressStr &progress)
{
    editProgressMsg("Verifying files existence...", progress);
    for(size_t i = 0; i < pkgList.size(); i++)
    {
        editProgress(float(i) / float(pkgList.size()) * 100.f, progress);

        string listFilename = "/var/lib/dpkg/info/" + pkgList[i] + ".list";

        ifstream file(listFilename.c_str(), ios::in);  // on ouvre en lecture

        if(file)  // si l'ouverture a fonctionné
        {
            string fileToCheck;
            while(getline(file, fileToCheck))
            {
                bool ok = true;

                if(access(fileToCheck.c_str(), F_OK) != 0)
                {
                    // Link ?
                    char tmp[1];
                    if(readlink(fileToCheck.c_str(), tmp, 1) == -1)
                    {
                        ok = false;
                        errList[i].nonExistingFilesList.push_back(fileToCheck);
                    }
                }

                if(progress.verboseMode == VERBOSE_MODE_VERBOSE)
                {
                    if(ok)
                    {
                        cout << "CheckFile - Ok : \"" << fileToCheck << "\"" << endl;
                    }
                    else
                    {
                        cout << "CheckFile - Error : \"" << fileToCheck << "\"" << endl;
                    }
                }
            }

            file.close();
            errList[i].listExists = true;
        }
        else
        {
            if(progress.verboseMode == VERBOSE_MODE_VERBOSE)
            {
                cout << "File \"" << listFilename << "\" doesn't exist !" << endl;
            }
            errList[i].listExists = false;
            continue;
        }
    }

    editProgress(100, progress);
}
//------------------------------------------------------------------------------
string getFileMD5(const string &filename)
{
    const size_t md5SumSize = 32;
    string md5 = "";

    string command = "md5sum \"" + filename + "\" 2> /dev/null";
    FILE *pipe = popen(command.c_str(), "r");

    if(pipe)
    {
        string tmp;

        if(getLine(tmp, pipe))
        {
            md5 = tmp.substr(0, md5SumSize);
        }

        pclose(pipe);
    }

    return md5;
}
//------------------------------------------------------------------------------
void checkMD5Sums(vector<errListStr> &errList, const vector<string> &pkgList, progressStr &progress)
{
    editProgressMsg("Verifying MD5 sums...", progress);

    const size_t md5SumSize = 32;
    const size_t fileNamePos = md5SumSize + 2;

    for(size_t i = 0; i < pkgList.size(); i++)
    {
        editProgress(float(i) / float(pkgList.size()) * 100.f, progress);

        string listFilename = "/var/lib/dpkg/info/" + pkgList[i] + ".md5sums";

        ifstream file(listFilename.c_str(), ios::in);  // on ouvre en lecture

        if(file)  // si l'ouverture a fonctionné
        {
            string line;
            string fileToCheck;
            string md5;
            string realMd5;
            while(getline(file, line))
            {
                md5 = line.substr(0, md5SumSize);
                fileToCheck = "/" + line.substr(fileNamePos);
                realMd5 = getFileMD5(fileToCheck);

                if(realMd5 != md5)
                {
                    if(progress.verboseMode == VERBOSE_MODE_VERBOSE)
                    {
                        cout << "CheckMD5 - Error : \"" << fileToCheck << "\"" << endl;
                    }

                    errList[i].md5errorList.push_back(fileToCheck);
                }
                else if(progress.verboseMode == VERBOSE_MODE_VERBOSE)
                {
                    cout << "CheckMD5 - Ok : \"" << fileToCheck << "\"" << endl;
                }
            }

            file.close();
            errList[i].md5Exists = true;
        }
        else
        {
            if(progress.verboseMode == VERBOSE_MODE_VERBOSE)
            {
                cout << "File \"" << listFilename << "\" doesn't exist !" << endl;
            }
            errList[i].md5Exists = true; // Possible pour quelques paquets ==> Tjr true.
            continue;
        }
    }

    editProgress(100, progress);
}

//------------------------------------------------------------------------------
int main(int argc, char *argv[])
{
    progressStr progress;
    pthread_mutex_init(&progress.mutex, NULL);
    progress.perCent = 0;
    progress.verboseMode = VERBOSE_MODE_NORMAL;
    progress.isInteractive = true;
    progress.isReinstall = true;
    progress.isMD5 = true;

    // Args...
    for(int i = 1; i < argc; i++)
    {
        string arg = argv[i];

        if(arg == "-v" || arg == "--verbose")
        {
            progress.verboseMode = VERBOSE_MODE_VERBOSE;
        }
        else if(arg == "-q" || arg == "--quiet")
        {
            progress.verboseMode = VERBOSE_MODE_QUIET;
        }
        else if(arg == "-a" || arg == "--auto")
        {
            progress.isInteractive = false;
        }
        else if(arg == "--no-md5")
        {
            progress.isMD5 = false;
        }
        else if(arg == "--no-reinstall")
        {
            progress.isReinstall = false;
        }
        else if(arg == "-V" || arg == "--version")
        {
            cout << "debtest " << VERSION << endl;
            return 0;
        }
        else if(arg == "-h" || arg == "--help")
        {
            cout << "Debtest will check your installed debian packages. It will verify files existency and md5sums. It can also reinstall corrupted packages." << endl;
            cout << "General:" << endl;
            cout << "  -V,  --version           Print current version." << endl;
            cout << "  -h,  --help              Show help message." << endl;
            cout << endl;
            cout << "Interaction:" << endl;
            cout << "  -q,  --quiet             Don't show progress." << endl;
            cout << "  -v,  --verbose           Verbose mode." << endl;
            cout << "  -a,  --auto              Non-interactive mode, packages are reinstalled automatically." << endl;
            cout << "  --no-reinstall           Don't reinstall any packages." << endl;
            cout << "  --no-md5                 Don't check md5 sums." << endl;
            return 0;
        }
        else
        {
            cerr << "Error : argument \"" << arg << "\" unknown." << endl;
            return 1;
        }
    }

    // Progress thread
    pthread_t progressThread;
    if(progress.verboseMode != VERBOSE_MODE_QUIET)
    {
        pthread_create(&progressThread, NULL, progressFn, (void*)&progress);
    }

    // Packages list
    editProgressMsg("Creating package list...", progress);
    vector<string> pkgList = getPkgList();
    vector<errListStr> errList;
    errList.resize(pkgList.size());

    // Check Files Existence
    checkFilesExistence(errList, pkgList, progress);

    // Check MD5 Sums
    if(progress.isMD5)
    {
        checkMD5Sums(errList, pkgList, progress);
    }

    // Recherche des paquets problématiques...
    vector<size_t> badPkgIndexes;
    for(size_t i = 0; i < errList.size(); i++)
    {
        // Si un fichier manquant...
        if(!errList[i].nonExistingFilesList.empty() || !errList[i].listExists)
        {
            badPkgIndexes.push_back(i);
        }
        else if(progress.isMD5) // Si non, on regarde du côté des MD5 (si besoin)
        {
            if(!errList[i].md5errorList.empty() || !errList[i].md5Exists)
            {
                badPkgIndexes.push_back(i);
            }
        }
    }

    // Saut de ligne + arret
    if(progress.verboseMode != VERBOSE_MODE_QUIET)
    {
        cout << endl;
        if(progress.verboseMode != VERBOSE_MODE_QUIET)
        {
            pthread_cancel(progressThread);
        }
    }

    // Info
    if( ! (progress.verboseMode == VERBOSE_MODE_QUIET && !progress.isInteractive))
    {
        if(badPkgIndexes.empty())
        {
            cout << "Everything is ok !" << endl;
        }
        else
        {
            cout << badPkgIndexes.size() << " packages are corrupted:"<< endl;
            for(size_t i = 0; i < badPkgIndexes.size(); i++)
            {
                cout << pkgList[badPkgIndexes[i]] << " ";
            }
            cout << endl << endl;
        }
    }

    if(badPkgIndexes.empty())
    {
        return 0;
    }

    // More infos ?
    if(progress.isInteractive)
    {
        cout << "Do you want more information about these packages ? (Y/n) " << flush;
        char letter;
        cin >> letter;

        if( ! (letter == 'n' || letter == 'N'))
        {
            for(size_t i = 0; i < badPkgIndexes.size(); i++)
            {
                size_t pkgId = badPkgIndexes[i];
                cout << endl << "== Package \"" << pkgList[pkgId] << "\" ==" << endl;
                if(!errList[pkgId].listExists)
                {
                    cout << "  Can't acces to file list !" << endl;
                }
                else if(!errList[pkgId].nonExistingFilesList.empty())
                {
                    for(size_t j = 0; j < errList[pkgId].nonExistingFilesList.size(); j++)
                    {
                        cout << "  File \"" << errList[pkgId].nonExistingFilesList[j] << "\" not found !" << endl;
                    }
                }

                if(progress.isMD5)
                {
                    if(!errList[pkgId].md5Exists)
                    {
                        cout << "  Can't acces to md5 list !" << endl;
                    }
                    else if(!errList[pkgId].md5errorList.empty())
                    {
                        for(size_t j = 0; j < errList[pkgId].md5errorList.size(); j++)
                        {
                            cout << "  File \"" << errList[pkgId].md5errorList[j] << "\" is currupted !" << endl;
                        }
                    }
                }
            }

            cout << endl;
        }
    }

    // Reparation
    if(progress.isReinstall)
    {
        if(progress.isInteractive)
        {
            cout << "Do you want to reinstall these packages ? (Y/n) " << flush;
            char letter;
            cin >> letter;

            if(letter == 'n' || letter == 'N')
            {
                return 0;
            }
        }

        string command = "apt-get install --reinstall -y ";
        for(size_t i = 0; i < badPkgIndexes.size(); i++)
        {
            command += pkgList[badPkgIndexes[i]] + " ";
        }

        if(system(command.c_str()) != 0)
        {
            cerr << "Error while reinstalling packages !" << endl;
        }
    }

    return 0;
}
